import os
import warnings
import pandas as pd

try:
    import fastavro as avro

    avro_imported = True
except ImportError:
    avro_imported = False


def read_df(path, **read_args):
    """Read a dataframe path based on the file extension
    parquet, avro, csv, pickle, json

    Parameters
    ----------
    path: str
        The path to the file holding data
    read_args : optional
        All keyword args are passed to the read function

    Returns
    -------
    data : pd.DataFrame

    Notes
    -----
    The read functions are taken from pandas, e.g. pd.read_csv
    Check the pandas doc for more information on the supported arguments
    """
    filetype = _get_extension(path)
    reader = _readers[filetype]
    if reader == pd.read_json:
        # Default json file is newline delimited json records, but can be overwritten
        defaults = {"lines": True, "orient": "records"}
        defaults.update(read_args)
        read_args = defaults

    return reader(path, **read_args)


def write_df(df, path, **write_args):
    """Write a dataframe to file based on the file extension

    The following formats are supported:
    parquet, avro, csv, pickle, json

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to write to disk
    datafile : DataFile
        Datafile instance with the path and file handle
    write_args : optional
        All keyword args are passed to the write function

    Notes
    -----
    The write functions are taken from pandas, e.g. pd.to_csv
    Check the pandas doc for more information on the supported arguments
    """
    extension = _get_extension(path)
    write_name = _writers[extension]

    # Some customizations for different file types
    if write_name == "to_avro":
        return _write_avro(df, path, **write_args)

    if write_name == "to_parquet" and not pd.Series(df.columns).map(type).eq(str).all():
        warnings.warn(
            "Dataframe contains non-string column names, which cannot be saved in parquet.\n"
            "Blocks will attempt to convert them to strings."
        )
        df.columns = df.columns.astype("str")

    if write_name == "to_json":
        defaults = {"lines": True, "orient": "records"}
        defaults.update(write_args)
        write_args = defaults

    if write_name == "to_csv":
        # make index=False the default for similar behaviour to other formats
        write_args["index"] = write_args.get("index", False)

    write_fn = getattr(df, write_name)
    write_fn(path, **write_args)


def _read_avro(path, **read_args):
    if not avro_imported:
        raise ImportError(
            "Avro support requires fastavro.\n"
            "Install blocks with the [avro] option or `pip install fastavro`"
        )
    records = []
    with open(path, "rb") as f:
        avro_reader = avro.reader(f)
        for record in avro_reader:
            records.append(record)
    return pd.DataFrame.from_dict(records)


def _write_avro(df, path, **write_args):
    if not avro_imported:
        raise ImportError(
            "Avro support requires fastavro.\n"
            "Install blocks with the [avro] option or `pip install fastavro`"
        )
    schema = None
    schema_path = None
    try:
        schema = write_args["schema"]
    except KeyError:
        try:
            schema_path = write_args["schema_path"]
        except KeyError:
            raise Exception(
                "You must provide a schema or schema path when writing to Avro"
            )
    if schema is None:
        schema = avro.schema.load_schema(schema_path)
    records = df.to_dict("records")
    with open(path, "wb") as f:
        avro.writer(f, schema, records)


def _get_extension(path):
    name, ext = os.path.splitext(path)
    return ext


_readers = {
    ".pq": pd.read_parquet,
    ".parquet": pd.read_parquet,
    ".csv": pd.read_csv,
    ".pkl": pd.read_pickle,
    ".avro": _read_avro,
    ".json": pd.read_json,
}


_writers = {
    ".pq": "to_parquet",
    ".parquet": "to_parquet",
    ".csv": "to_csv",
    ".pkl": "to_pickle",
    ".avro": "to_avro",
    ".json": "to_json",
}



from blocks.filesystem.gcs_filesystem import GCSFileSystem  # NOQA
from blocks.filesystem.gcs_native_filesystem import GCSNativeFileSystem  # NOQA
from blocks.filesystem.base import FileSystem  # NOQA



import warnings

from blocks.filesystem.base import FileSystem


class GCSFileSystem(FileSystem):
    """Deprecated, see FileSystem"""

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "This class is deprecated, use blocks.filesystem.FileSystem",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)



import atexit
import tempfile

import wrapt


@wrapt.decorator
def with_function_tmpdir(wrapped, instance, args, kwargs):
    with tempfile.TemporaryDirectory() as tmpdir:
        kwargs["tmpdir"] = tmpdir
        return wrapped(*args, **kwargs)


@wrapt.decorator
def with_session_tmpdir(wrapped, instance, args, kwargs):
    tmpdir = tempfile.TemporaryDirectory()
    kwargs["tmpdir"] = tmpdir.name
    atexit.register(tmpdir.cleanup)
    return wrapped(*args, **kwargs)



from typing import Union, Sequence, Tuple, List
from collections import defaultdict
from fsspec.core import split_protocol, get_filesystem_class, has_magic


class FileSystem:
    """Filesystem for manipulating files in the cloud

    This supports operations on local files and any other protocol supported by fsspec.
    This is a wrapper to fsspec which provides backwards compatibility for blocks filesystems
    and a simplified interface.

    Parameters
    ----------
    storage_options: Mapping[str, Mapping[str, Any]]
        Additional options passed to each filesystem for each protocol
        e.g. {'gs': {'project': 'example'}} to set the gs filesytem project to example
    """

    def __init__(self, **storage_options):
        self.storage_options = defaultdict(dict)
        self.storage_options.update(storage_options)
        self.storage_options[None]["auto_mkdir"] = True
        self.filesystems = {}

    def _get_protocol_path(self, urlpath) -> Tuple[str, List[str]]:
        if isinstance(urlpath, str):
            return split_protocol(urlpath)

        protocols, paths = zip(*map(split_protocol, urlpath))
        assert (
            len(set(protocols)) == 1
        ), "Cannot mix file protocols in a single operation"
        return protocols[0], list(paths)

    def _get_filesystem(self, protocol):
        if protocol not in self.filesystems:
            self.filesystems[protocol] = get_filesystem_class(protocol)(
                **self.storage_options[protocol]
            )
        return self.filesystems[protocol]

    def ls(self, path: str) -> Sequence[str]:
        """List files correspond to path, including glob wildcards

        Parameters
        ----------
        path : str
            The path to the file or directory to list; supports wildcards
        """
        protocol, path = self._get_protocol_path(path)
        fs = self._get_filesystem(protocol)
        try:
            if has_magic(path):
                output = fs.glob(path)
            else:
                output = fs.ls(path)
        # TODO fix in base
        except FileNotFoundError:
            return []
        except NotADirectoryError:
            return [path]

        if protocol is not None:
            output = ["://".join([protocol, path]) for path in output]
        return sorted(output)

    def copy(
        self,
        sources: Union[str, Sequence[str]],
        dest: Union[str, Sequence[str]],
        recursive=False,
    ):
        """Copy the files in sources to dest

        Parameters
        ----------
        sources : list of str
            The list of paths to copy
        dest : str
            The destination(s) for the copy of source(s)
        recursive : bool
            If true, recursively copy any directories
        """
        if isinstance(sources, str):
            sources = [sources]

        protocol_source, sources = self._get_protocol_path(sources)
        protocol_dest, dest = self._get_protocol_path(dest)

        if protocol_source == protocol_dest:
            fs = self._get_filesystem(protocol_source)

            # Temporary workaround for a bug in gcsfs
            if protocol_source == "gs" and recursive:
                sources = fs.expand_path(sources, recursive=True)
                sources = ["gs://" + s for s in sources if not fs.isdir(s)]
                return self.copy(sources, "gs://" + dest, recursive=False)

            fs.copy(sources, dest, recursive=recursive)

        elif protocol_source is None:
            fs = self._get_filesystem(protocol_dest)
            fs.put(sources, dest, recursive=recursive)

        elif protocol_dest is None:
            fs = self._get_filesystem(protocol_source)
            fs.get(sources, dest, recursive=recursive)

        elif protocol_dest is not None and protocol_source is not None:
            raise NotImplementedError(
                "Cannot do direct copy between two different cloud filesystems"
            )

        if protocol_dest == "gs":
            # Make sure we invalidate the gcsfs cache since we have added new files
            if isinstance(dest, str):
                fs.invalidate_cache(dest)
            else:
                for d in dest:
                    fs.invalidate_cache(d)

    def remove(self, paths: Union[str, List[str]], recursive: bool = False):
        """Remove the files at paths

        Parameters
        ----------
        paths : list of str
            The paths to remove
        recursive : bool, default False
            If true, recursively remove any directories
        """
        protocol, paths = self._get_protocol_path(paths)
        fs = self._get_filesystem(protocol)

        if protocol is None and not isinstance(paths, str):
            # TODO should local not just handle this?
            for path in paths:
                fs.rm(path, recursive=recursive)
        else:
            return fs.rm(paths, recursive=recursive)

    def open(self, path: str, mode="rb", **kwargs):
        """Return a file-like object from the filesystem

        The resultant instance must function correctly in a context ``with``
        block.

        Parameters
        ----------
        path: str
            Target file
        mode: str like 'rb', 'w'
            See builtin ``open()``
        kwargs:
            Forwarded to the filesystem implementation
        """
        protocol, path = self._get_protocol_path(path)
        fs = self._get_filesystem(protocol)
        return fs.open(path, mode, **kwargs)

    def isdir(self, path: str):
        """Check if the path is a directory"""
        protocol, path = self._get_protocol_path(path)
        fs = self._get_filesystem(protocol)
        return fs.isdir(path)

    def mkdir(self, path: str):
        """Make directory at path"""
        protocol, path = self._get_protocol_path(path)
        fs = self._get_filesystem(protocol)
        return fs.mkdir(path)

    # Aliases
    cp = copy
    rm = remove
